@@ -2,16 +2,19 @@ import os.path
 import sys
 import traceback
 from collections import OrderedDict, defaultdict
-from contextlib import contextmanager
 
-from typing import Tuple, List, TypeVar, Set, Dict, Iterator, Optional, cast
+from typing import Tuple, List, TypeVar, Set, Dict, Optional
 
+from mypy.scope import Scope
 from mypy.options import Options
 from mypy.version import __version__ as mypy_version
 
+MYPY = False
+if MYPY:
+    from typing_extensions import Final
 
 T = TypeVar('T')
-allowed_duplicates = ['@overload', 'Got:', 'Expected:']
+allowed_duplicates = ['@overload', 'Got:', 'Expected:']  # type: Final
 
 
 class ErrorInfo:
@@ -21,7 +24,7 @@ class ErrorInfo:
     # related to this error. Each item is a (path, line number) tuple.
     import_ctx = None  # type: List[Tuple[str, int]]
 
-    # The source file that was the source of this error.
+    # The path to source file that was the source of this error.
     file = ''
 
     # The fully-qualified id of the source module for this error.
@@ -51,6 +54,9 @@ class ErrorInfo:
     # Only report this particular messages once per program.
     only_once = False
 
+    # Actual origin of the error message as tuple (path, line number)
+    origin = None  # type: Tuple[str, int]
+
     # Fine-grained incremental target where this was reported
     target = None  # type: Optional[str]
 
@@ -90,26 +96,22 @@ class Errors:
     current error context (nested imports).
     """
 
-    # List of generated error messages.
-    error_info = None  # type: List[ErrorInfo]
+    # Map from files to generated error messages. Is an OrderedDict so
+    # that it can be used to order messages based on the order the
+    # files were processed.
+    error_info_map = None  # type: Dict[str, List[ErrorInfo]]
+
+    # Files that we have reported the errors for
+    flushed_files = None  # type: Set[str]
 
     # Current error context: nested import context/stack, as a list of (path, line) pairs.
     import_ctx = None  # type: List[Tuple[str, int]]
 
-    # Set of files with errors.
-    error_files = None  # type: Set[str]
-
     # Path name prefix that is removed from all paths, if set.
-    ignore_prefix = None  # type: str
+    ignore_prefix = None  # type: Optional[str]
 
     # Path to current file.
-    file = None  # type: str
-
-    # Stack of short names of currents types (or None).
-    type_name = None  # type: List[Optional[str]]
-
-    # Stack of short names of current functions or members (or None).
-    function_or_member = None  # type: List[Optional[str]]
+    file = ''  # type: str
 
     # Ignore errors on these lines of each file.
     ignored_lines = None  # type: Dict[str, Set[int]]
@@ -129,10 +131,11 @@ class Errors:
     # Set to True to show column numbers in error messages.
     show_column_numbers = False  # type: bool
 
-    # Stack of active fine-grained incremental checking targets within
-    # a module. The first item is always the current module id.
+    # State for keeping track of the current fine-grained incremental mode target.
     # (See mypy.server.update for more about targets.)
-    target = None  # type: List[str]
+    # Current module id.
+    target_module = None  # type: Optional[str]
+    scope = None  # type: Optional[Scope]
 
     def __init__(self, show_error_context: bool = False,
                  show_column_numbers: bool = False) -> None:
@@ -141,16 +144,17 @@ class Errors:
         self.initialize()
 
     def initialize(self) -> None:
-        self.error_info = []
+        self.error_info_map = OrderedDict()
+        self.flushed_files = set()
         self.import_ctx = []
-        self.error_files = set()
         self.type_name = [None]
         self.function_or_member = [None]
         self.ignored_lines = OrderedDict()
         self.used_ignored_lines = defaultdict(set)
         self.ignored_files = set()
         self.only_once_messages = set()
-        self.target = []
+        self.scope = None
+        self.target_module = None
 
     def reset(self) -> None:
         self.initialize()
@@ -161,9 +165,13 @@ class Errors:
         new.import_ctx = self.import_ctx[:]
         new.type_name = self.type_name[:]
         new.function_or_member = self.function_or_member[:]
-        new.target = self.target[:]
+        new.target_module = self.target_module
+        new.scope = self.scope
         return new
 
+    def total_errors(self) -> int:
+        return sum(len(errs) for errs in self.error_info_map.values())
+
     def set_ignore_prefix(self, prefix: str) -> None:
         """Set path prefix that will be removed from all paths."""
         prefix = os.path.normpath(prefix)
@@ -178,7 +186,7 @@ class Errors:
 
     def set_file(self, file: str,
                  module: Optional[str],
-                 ignored_lines: Optional[Set[int]] = None) -> None:
+                 scope: Optional[Scope] = None) -> None:
         """Set the path and module id of the current file."""
         # The path will be simplified later, in render_messages. That way
         #  * 'file' is always a key that uniquely identifies a source file
@@ -187,8 +195,8 @@ class Errors:
         #    reporting errors for files other than the one currently being
         #    processed.
         self.file = file
-        if module:
-            self.target = [module]
+        self.target_module = module
+        self.scope = scope
 
     def set_file_ignored_lines(self, file: str,
                                ignored_lines: Set[int],
@@ -197,54 +205,16 @@ class Errors:
         if ignore_all:
             self.ignored_files.add(file)
 
-    def push_function(self, name: str) -> None:
-        """Set the current function or member short name (it can be None)."""
-        self.push_target_component(name)
-        self.function_or_member.append(name)
-
-    def pop_function(self) -> None:
-        self.function_or_member.pop()
-        self.pop_target_component()
-
-    @contextmanager
-    def enter_function(self, name: str) -> Iterator[None]:
-        self.push_function(name)
-        yield
-        self.pop_function()
-
-    def push_type(self, name: str) -> None:
-        """Set the short name of the current type (it can be None)."""
-        self.push_target_component(name)
-        self.type_name.append(name)
-
-    def pop_type(self) -> None:
-        self.type_name.pop()
-        self.pop_target_component()
-
-    def push_target_component(self, name: str) -> None:
-        if self.target and not self.function_or_member[-1]:
-            self.target.append('{}.{}'.format(self.target[-1], name))
-
-    def pop_target_component(self) -> None:
-        if self.target and not self.function_or_member[-1]:
-            self.target.pop()
-
     def current_target(self) -> Optional[str]:
-        if self.target:
-            return self.target[-1]
-        return None
+        """Retrieves the current target from the associated scope.
+
+        If there is no associated scope, use the target module."""
+        if self.scope is not None:
+            return self.scope.current_target()
+        return self.target_module
 
     def current_module(self) -> Optional[str]:
-        if self.target:
-            return self.target[0]
-        return None
-
-    @contextmanager
-    def enter_type(self, name: str) -> Iterator[None]:
-        """Set the short name of the current type (it can be None)."""
-        self.push_type(name)
-        yield
-        self.pop_type()
+        return self.target_module
 
     def import_context(self) -> List[Tuple[str, int]]:
         """Return a copy of the import context."""
@@ -256,7 +226,7 @@ class Errors:
 
     def report(self,
                line: int,
-               column: int,
+               column: Optional[int],
                message: str,
                blocker: bool = False,
                severity: str = 'error',
@@ -275,22 +245,36 @@ class Errors:
             only_once: if True, only report this exact message once per build
             origin_line: if non-None, override current context as origin
         """
-        type = self.type_name[-1]  # type: Optional[str]
-        if len(self.function_or_member) > 2:
-            type = None  # Omit type context if nested function
+        if self.scope:
+            type = self.scope.current_type_name()
+            if self.scope.ignored > 0:
+                type = None  # Omit type context if nested function
+            function = self.scope.current_function_name()
+        else:
+            type = None
+            function = None
+
+        if column is None:
+            column = -1
         if file is None:
             file = self.file
         if offset:
             message = " " * offset + message
         info = ErrorInfo(self.import_context(), file, self.current_module(), type,
-                         self.function_or_member[-1], line, column, severity, message,
+                         function, line, column, severity, message,
                          blocker, only_once,
                          origin=(self.file, origin_line) if origin_line else None,
                          target=self.current_target())
         self.add_error_info(info)
 
+    def _add_error_info(self, file: str, info: ErrorInfo) -> None:
+        assert file not in self.flushed_files
+        if file not in self.error_info_map:
+            self.error_info_map[file] = []
+        self.error_info_map[file].append(info)
+
     def add_error_info(self, info: ErrorInfo) -> None:
-        (file, line) = cast(Tuple[str, int], info.origin)  # see issue 1855
+        file, line = info.origin
         if not info.blocker:  # Blockers cannot be ignored
             if file in self.ignored_lines and line in self.ignored_lines[file]:
                 # Annotation requests us to ignore all errors on this line.
@@ -302,18 +286,28 @@ class Errors:
             if info.message in self.only_once_messages:
                 return
             self.only_once_messages.add(info.message)
-        self.error_info.append(info)
-        self.error_files.add(file)
+        self._add_error_info(file, info)
 
-    def generate_unused_ignore_notes(self) -> None:
-        for file, ignored_lines in self.ignored_lines.items():
-            if not self.is_typeshed_file(file):
-                for line in ignored_lines - self.used_ignored_lines[file]:
-                    # Don't use report since add_error_info will ignore the error!
-                    info = ErrorInfo(self.import_context(), file, self.current_module(), None,
-                                     None, line, -1, 'note', "unused 'type: ignore' comment",
-                                     False, False)
-                    self.error_info.append(info)
+    def clear_errors_in_targets(self, path: str, targets: Set[str]) -> None:
+        """Remove errors in specific fine-grained targets within a file."""
+        if path in self.error_info_map:
+            new_errors = []
+            for info in self.error_info_map[path]:
+                if info.target not in targets:
+                    new_errors.append(info)
+                elif info.only_once:
+                    self.only_once_messages.remove(info.message)
+            self.error_info_map[path] = new_errors
+
+    def generate_unused_ignore_notes(self, file: str) -> None:
+        ignored_lines = self.ignored_lines[file]
+        if not self.is_typeshed_file(file) and file not in self.ignored_files:
+            for line in ignored_lines - self.used_ignored_lines[file]:
+                # Don't use report since add_error_info will ignore the error!
+                info = ErrorInfo(self.import_context(), file, self.current_module(), None,
+                                 None, line, -1, 'note', "unused 'type: ignore' comment",
+                                 False, False)
+                self._add_error_info(file, info)
 
     def is_typeshed_file(self, file: str) -> bool:
         # gross, but no other clear way to tell
@@ -321,50 +315,53 @@ class Errors:
 
     def num_messages(self) -> int:
         """Return the number of generated messages."""
-        return len(self.error_info)
+        return sum(len(x) for x in self.error_info_map.values())
 
     def is_errors(self) -> bool:
         """Are there any generated errors?"""
-        return bool(self.error_info)
+        return bool(self.error_info_map)
 
     def is_blockers(self) -> bool:
         """Are the any errors that are blockers?"""
-        return any(err for err in self.error_info if err.blocker)
+        return any(err for errs in self.error_info_map.values() for err in errs if err.blocker)
 
     def blocker_module(self) -> Optional[str]:
         """Return the module with a blocking error, or None if not possible."""
-        for err in self.error_info:
-            if err.blocker:
-                return err.module
+        for errs in self.error_info_map.values():
+            for err in errs:
+                if err.blocker:
+                    return err.module
         return None
 
     def is_errors_for_file(self, file: str) -> bool:
         """Are there any errors for the given file?"""
-        return file in self.error_files
+        return file in self.error_info_map
 
     def raise_error(self) -> None:
         """Raise a CompileError with the generated messages.
 
         Render the messages suitable for displaying.
         """
-        raise CompileError(self.messages(),
+        # self.new_messages() will format all messages that haven't already
+        # been returned from a file_messages() call.
+        raise CompileError(self.new_messages(),
                            use_stdout=True,
                            module_with_blocker=self.blocker_module())
 
-    def messages(self) -> List[str]:
+    def format_messages(self, error_info: List[ErrorInfo]) -> List[str]:
         """Return a string list that represents the error messages.
 
         Use a form suitable for displaying to the user.
         """
         a = []  # type: List[str]
-        errors = self.render_messages(self.sort_messages(self.error_info))
+        errors = self.render_messages(self.sort_messages(error_info))
         errors = self.remove_duplicates(errors)
         for file, line, column, severity, message in errors:
             s = ''
             if file is not None:
                 if self.show_column_numbers and line is not None and line >= 0 \
                         and column is not None and column >= 0:
-                    srcloc = '{}:{}:{}'.format(file, line, column)
+                    srcloc = '{}:{}:{}'.format(file, line, 1 + column)
                 elif line is not None and line >= 0:
                     srcloc = '{}:{}'.format(file, line)
                 else:
@@ -375,22 +372,46 @@ class Errors:
             a.append(s)
         return a
 
+    def file_messages(self, path: str) -> List[str]:
+        """Return a string list of new error messages from a given file.
+
+        Use a form suitable for displaying to the user.
+        """
+        if path not in self.error_info_map:
+            return []
+        self.flushed_files.add(path)
+        return self.format_messages(self.error_info_map[path])
+
+    def new_messages(self) -> List[str]:
+        """Return a string list of new error messages.
+
+        Use a form suitable for displaying to the user.
+        Errors from different files are ordered based on the order in which
+        they first generated an error.
+        """
+        msgs = []
+        for path in self.error_info_map.keys():
+            if path not in self.flushed_files:
+                msgs.extend(self.file_messages(path))
+        return msgs
+
     def targets(self) -> Set[str]:
         """Return a set of all targets that contain errors."""
         # TODO: Make sure that either target is always defined or that not being defined
         #       is okay for fine-grained incremental checking.
         return set(info.target
-                   for info in self.error_info
+                   for errs in self.error_info_map.values()
+                   for info in errs
                    if info.target)
 
     def render_messages(self, errors: List[ErrorInfo]) -> List[Tuple[Optional[str], int, int,
                                                                      str, str]]:
         """Translate the messages into a sequence of tuples.
 
-        Each tuple is of form (path, line, col, message.  The rendered
-        sequence includes information about error contexts. The path
-        item may be None. If the line item is negative, the line
-        number is not defined for the tuple.
+        Each tuple is of form (path, line, col, severity, message).
+        The rendered sequence includes information about error contexts.
+        The path item may be None. If the line item is negative, the
+        line number is not defined for the tuple.
         """
         result = []  # type: List[Tuple[Optional[str], int, int, str, str]]
         # (path, line, column, severity, message)
@@ -461,7 +482,7 @@ class Errors:
     def sort_messages(self, errors: List[ErrorInfo]) -> List[ErrorInfo]:
         """Sort an array of error messages locally by line number.
 
-        I.e., sort a run of consecutive messages with the same file
+        I.e., sort a run of consecutive messages with the same
         context by line number, but otherwise retain the general
         ordering of the messages.
         """
@@ -492,7 +513,7 @@ class Errors:
             while (j >= 0 and errors[j][0] == errors[i][0] and
                     errors[j][1] == errors[i][1]):
                 if (errors[j][3] == errors[i][3] and
-                        # Allow duplicate notes in overload conficts reporting
+                        # Allow duplicate notes in overload conflicts reporting.
                         not (errors[i][3] == 'note' and
                              errors[i][4].strip() in allowed_duplicates
                              or errors[i][4].strip().startswith('def ')) and
@@ -511,6 +532,12 @@ class CompileError(Exception):
 
     It can be a parse, semantic analysis, type check or other
     compilation-related error.
+
+    CompileErrors raised from an errors object carry all of the
+    messages that have not been reported out by error streaming.
+    This is patched up by build.build to contain either all error
+    messages (if errors were streamed) or none (if they were not).
+
     """
 
     messages = None  # type: List[str]
@@ -528,14 +555,7 @@ class CompileError(Exception):
         self.module_with_blocker = module_with_blocker
 
 
-class DecodeError(Exception):
-    """Exception raised when a file cannot be decoded due to an unknown encoding type.
-
-    Essentially a wrapper for the LookupError raised by `bytearray.decode`
-    """
-
-
-def remove_path_prefix(path: str, prefix: str) -> str:
+def remove_path_prefix(path: str, prefix: Optional[str]) -> str:
     """If path starts with prefix, return copy of path with the prefix removed.
     Otherwise, return path. If path is None, return None.
     """
@@ -554,7 +574,7 @@ def report_internal_error(err: Exception, file: Optional[str], line: int,
     # Dump out errors so far, they often provide a clue.
     # But catch unexpected errors rendering them.
     try:
-        for msg in errors.messages():
+        for msg in errors.new_messages():
             print(msg)
     except Exception as e:
         print("Failed to dump errors:", repr(e), file=sys.stderr)
@@ -581,6 +601,8 @@ def report_internal_error(err: Exception, file: Optional[str], line: int,
         pdb.post_mortem(sys.exc_info()[2])
 
     # If requested, print traceback, else print note explaining how to get one.
+    if options.raise_exceptions:
+        raise err
     if not options.show_traceback:
         if not options.pdb:
             print('{}: note: please use --show-traceback to print a traceback '
@@ -596,4 +618,5 @@ def report_internal_error(err: Exception, file: Optional[str], line: int,
         print('{}: note: use --pdb to drop into pdb'.format(prefix), file=sys.stderr)
 
     # Exit.  The caller has nothing more to say.
-    raise SystemExit(1)
+    # We use exit code 2 to signal that this is no ordinary error.
+    raise SystemExit(2)
