@@ -4,7 +4,7 @@ When the source code of a module has a change in fine-grained incremental mode,
 we build a new AST from the updated source. However, other parts of the program
 may have direct references to parts of the old AST (namely, those nodes exposed
 in the module symbol table). The merge operation changes the identities of new
-AST nodes that have a correspondance in the old AST to the old ones so that
+AST nodes that have a correspondence in the old AST to the old ones so that
 existing cross-references in other modules will continue to point to the correct
 nodes. Also internal cross-references within the new AST are replaced. AST nodes
 that aren't externally visible will get new, distinct object identities. This
@@ -15,7 +15,7 @@ external references (which would be slow and fragile) or always perform
 translation when looking up references (which would be hard to retrofit).
 
 The AST merge operation is performed after semantic analysis. Semantic
-analysis has to deal with potentionally multiple aliases to certain AST
+analysis has to deal with potentially multiple aliases to certain AST
 nodes (in particular, MypyFile nodes). Type checking assumes that we
 don't have multiple variants of a single AST node visible to the type
 checker.
@@ -28,7 +28,7 @@ Discussion of some notable special cases:
 
 * If a function is replaced with another function with an identical signature,
   call sites continue to point to the same object (by identity) and don't need
-  to be reprocessed. Similary, if a class is replaced with a class that is
+  to be reprocessed. Similarly, if a class is replaced with a class that is
   sufficiently similar (MRO preserved, etc.), class references don't need any
   processing. A typical incremental update to a file only changes a few
   externally visible things in a module, and this means that often only few
@@ -48,7 +48,7 @@ See the main entry point merge_asts for more details.
 from typing import Dict, List, cast, TypeVar, Optional
 
 from mypy.nodes import (
-    Node, MypyFile, SymbolTable, Block, AssignmentStmt, NameExpr, MemberExpr, RefExpr, TypeInfo,
+    MypyFile, SymbolTable, Block, AssignmentStmt, NameExpr, MemberExpr, RefExpr, TypeInfo,
     FuncDef, ClassDef, NamedTupleExpr, SymbolNode, Var, Statement, SuperExpr, NewTypeExpr,
     OverloadedFuncDef, LambdaExpr, TypedDictExpr, EnumCallExpr, FuncBase, TypeAliasExpr, CallExpr,
     CastExpr,
@@ -58,9 +58,11 @@ from mypy.traverser import TraverserVisitor
 from mypy.types import (
     Type, SyntheticTypeVisitor, Instance, AnyType, NoneTyp, CallableType, DeletedType, PartialType,
     TupleType, TypeType, TypeVarType, TypedDictType, UnboundType, UninhabitedType, UnionType,
-    Overloaded, TypeVarDef, TypeList, CallableArgument, EllipsisType, StarType
+    Overloaded, TypeVarDef, TypeList, CallableArgument, EllipsisType, StarType, LiteralType,
+    RawExpressionType,
 )
-from mypy.util import get_prefix
+from mypy.util import get_prefix, replace_object_state
+from mypy.typestate import TypeState
 
 
 def merge_asts(old: MypyFile, old_symbols: SymbolTable,
@@ -255,7 +257,6 @@ class NodeReplaceVisitor(TraverserVisitor):
 
     def visit_type_alias_expr(self, node: TypeAliasExpr) -> None:
         self.fixup_type(node.type)
-        self.fixup_type(node.fallback)
         super().visit_type_alias_expr(node)
 
     # Others
@@ -270,7 +271,7 @@ class NodeReplaceVisitor(TraverserVisitor):
     def fixup(self, node: SN) -> SN:
         if node in self.replacements:
             new = self.replacements[node]
-            new.__dict__ = node.__dict__
+            replace_object_state(new, node)
             return cast(SN, new)
         return node
 
@@ -284,7 +285,7 @@ class NodeReplaceVisitor(TraverserVisitor):
             # The subclass relationships may change, so reset all caches relevant to the
             # old MRO.
             new = cast(TypeInfo, self.replacements[node])
-            new.reset_subtype_cache()
+            TypeState.reset_all_subtype_caches_for(new)
         return self.fixup(node)
 
     def fixup_type(self, typ: Optional[Type]) -> None:
@@ -300,7 +301,8 @@ class NodeReplaceVisitor(TraverserVisitor):
         self.fixup_type(info.tuple_type)
         self.fixup_type(info.typeddict_type)
         info.defn.info = self.fixup(info)
-        info.replaced = self.fixup(info.replaced)
+        if info.replaced:
+            info.replaced = self.fixup(info.replaced)
         replace_nodes_in_symbol_table(info.names, self.replacements)
         for i, item in enumerate(info.mro):
             info.mro[i] = self.fixup(info.mro[i])
@@ -326,7 +328,12 @@ class NodeReplaceVisitor(TraverserVisitor):
 
 
 class TypeReplaceVisitor(SyntheticTypeVisitor[None]):
-    """Similar to NodeReplaceVisitor, but for type objects."""
+    """Similar to NodeReplaceVisitor, but for type objects.
+
+    Note: this visitor may sometimes visit unanalyzed types
+    such as 'UnboundType' and 'RawExpressionType' For example, see
+    NodeReplaceVisitor.process_base_func.
+    """
 
     def __init__(self, replacements: Dict[SymbolNode, SymbolNode]) -> None:
         self.replacements = replacements
@@ -335,6 +342,8 @@ class TypeReplaceVisitor(SyntheticTypeVisitor[None]):
         typ.type = self.fixup(typ.type)
         for arg in typ.args:
             arg.accept(self)
+        if typ.final_value:
+            typ.final_value.accept(self)
 
     def visit_any(self, typ: AnyType) -> None:
         pass
@@ -374,8 +383,8 @@ class TypeReplaceVisitor(SyntheticTypeVisitor[None]):
         for item in typ.items:
             item.accept(self)
         # Fallback can be None for implicit tuple types that haven't been semantically analyzed.
-        if typ.fallback is not None:
-            typ.fallback.accept(self)
+        if typ.partial_fallback is not None:
+            typ.partial_fallback.accept(self)
 
     def visit_type_type(self, typ: TypeType) -> None:
         typ.item.accept(self)
@@ -390,6 +399,12 @@ class TypeReplaceVisitor(SyntheticTypeVisitor[None]):
             value_type.accept(self)
         typ.fallback.accept(self)
 
+    def visit_raw_expression_type(self, t: RawExpressionType) -> None:
+        pass
+
+    def visit_literal_type(self, typ: LiteralType) -> None:
+        typ.fallback.accept(self)
+
     def visit_unbound_type(self, typ: UnboundType) -> None:
         for arg in typ.args:
             arg.accept(self)
@@ -419,8 +434,6 @@ class TypeReplaceVisitor(SyntheticTypeVisitor[None]):
     def fixup(self, node: SN) -> SN:
         if node in self.replacements:
             new = self.replacements[node]
-            # TODO: This may be unnecessary?
-            new.__dict__ = node.__dict__
             return cast(SN, new)
         return node
 
@@ -431,15 +444,13 @@ def replace_nodes_in_symbol_table(symbols: SymbolTable,
         if node.node:
             if node.node in replacements:
                 new = replacements[node.node]
-                new.__dict__ = node.node.__dict__
+                old = node.node
+                replace_object_state(new, old)
                 node.node = new
             if isinstance(node.node, Var):
                 # Handle them here just in case these aren't exposed through the AST.
                 # TODO: Is this necessary?
                 fixup_var(node.node, replacements)
-        override = node.type_override
-        if override:
-            override.accept(TypeReplaceVisitor(replacements))
 
 
 def fixup_var(node: Var, replacements: Dict[SymbolNode, SymbolNode]) -> None:
